/* -*- mode: c++ -*- */
#ifndef __Machine_h__
#define __Machine_h__
#include <glib.h>
#ifdef HAVE_CONFIG_H
#ifdef G_OS_WIN32
#include "config_win.h"
#else
#include <config.h>
#endif
#else
#warning "NO config"
#endif

#include <DataBuffer.h>

/*
 * The data coming from a chip
 */
struct ChipBlock
{
        unsigned short header[16];
        unsigned short signal[128];
};
/*
 * This is the basic data chunk coming from USB
 */
struct EventDataBlock
{
        unsigned int time;
        unsigned short temp;
        ChipBlock chip[2];
};


class TheMachine
{
    public:
        enum MachineState
        {
            NotInitialized,
            SoftwareReset,
            WaitingForHandshake,
            Waiting,
            Initialized,
            ReadyForData,
            BeetleConfiguration,
            BeetleConfigReady,
            BeetleCalibration,
            AlibavaConf,
            AlibavaConfReady,
            AlibavaConfEnd,
            TriggerCount,
            TriggerCountReady,
            TriggerCountEnd,
            TriggerIn,
            TriggerDACs,
            LaserSynchronization,
            PrepareAcquisition,
            Acquiring,
            ReadingData,
            WaitReadAck,
            last_state
        };

        enum DAQtype
        {
            laser,
            source,
            pedestal,
            last_type
        };
    private:
        int fw_version;
        int chip_mask;
        int nreset;
        unsigned short ntrigger;
        unsigned char _state;
        MachineState _mstate;
        unsigned short _beetle_conf[22];
        unsigned short _fw_config;
        double _thrs; // Threshold in electrons
        int _error;
        int indx;
        DataBuffer _buffer;
        std::vector<EventDataBlock *> _memory;
        /// the handler for the idle process in the glib loop
        int _idle_pedestal;
        int _idle_signal;
        int _idle_send_data;

        // number of events required for an acquisition
        unsigned short _nevts;
        unsigned short _nsent;
        GIOChannel *_fd;
        DAQtype daq_type;

        double _pedestals[256];

        void *tmp_data;

        static int _generate_pedestals(TheMachine *);
        static int _generate_signal(TheMachine *);
        static int _send_signal(TheMachine *);
        static const char *sstate[];

        int get_chip_number();
    public:
        TheMachine();

        ~TheMachine()
        {
        }

        void init(GIOChannel *fd=0);
        bool need_init() const
        {
            return _mstate==NotInitialized;
        }

        int error() const
        {
            return _error;
        }
        void error(int x)
        {
            _error = x;
        }

        int get_chip_mask() const { return chip_mask; }

        void set_fw_version(int x);
        int get_fw_version() const { return fw_version; }

        unsigned char read_command(GIOChannel *fd);

        unsigned short read_word(GIOChannel *fd);

        unsigned char state() const
        {
            return _state;
        }
        void state(unsigned char s)
        {
            _state = s;
        }
        MachineState mstate() const
        {
            return _mstate;
        }
        void mstate(MachineState s);

        void hand_shake(GIOChannel *);

        void set_to_wait(GIOChannel *fd);
        void send_state(GIOChannel *, unsigned char);

        int process_state(GIOChannel *);
        int send_data_block(GIOChannel *, int);

        void create_data();
        DataBlock *get_data();

        void reset();
        int beetle_config(GIOChannel *);
        int beetle_calib(GIOChannel *);
        int trigger_config(GIOChannel *);
        int laser_synchronization(GIOChannel *);

        int get_nbits();
};

#endif
