#include "PythonLoader.h"
#include <iostream>
#include <sstream>
#include <vector>
#include <map>
#include <string>
#include <glib.h>
#include "mingw.h"
#include "utils.h"

#if defined(HAS_PYTHON)

std::map<std::string, int> __search_path__;

void init_python_objects();

PythonLoader PythonLoader::__the_loader;
#if (PY_VERSION_HEX < 0x02040000)
#define PyEval_ThreadsInitialized() (true)
#endif
/**
 * TODO: WHY !?!?!?!?
 * We need to understand why declaring _PyGC_generation0 here
 * stops the problem wirh
 * vmedaq: Modules/gcmodule.c:338:
 * visit_reachable: Assertion `gc_refs > 0 || gc_refs == (-3) || gc_refs == (-2)' failed
 *
 */
#define NUM_GENERATIONS 3
#define GEN_HEAD(n) (&generations[n].head)
struct gc_generation {
    PyGC_Head head;
    int threshold; /* collection threshold */
    int count; /* count of allocations or collections of younger
              generations */
};
/* linked lists of container objects */
static struct gc_generation generations[NUM_GENERATIONS] = {
    /* PyGC_Head,               threshold,  count */
    {{{GEN_HEAD(0), GEN_HEAD(0), 0}},   700,        0},
    {{{GEN_HEAD(1), GEN_HEAD(1), 0}},   10,     0},
    {{{GEN_HEAD(2), GEN_HEAD(2), 0}},   10,     0},
};

PyGC_Head *_PyGC_generation0 = GEN_HEAD(0);



/*
 * Get the name of the current function
 *
 */
std::string get_function_name(int level=2)
{
    std::string out;
    PyObject *inspect=0;
    PyObject *stack_func=0;
    PyObject *stack_list=0;
    PyObject *frame_record = 0;
    PyObject *frame = 0;
    PyObject *f_code = 0;
    PyObject *co_name = 0;
    int list_length = 0;

    inspect = PyImport_ImportModule("inspect");
    if (!inspect)
        goto done;

    stack_func = PyObject_GetAttrString(inspect, "stack");
    if (!stack_func)
        goto done;

    stack_list = PyObject_CallObject(stack_func, NULL);
    if (!stack_list)
        goto done;

    list_length = PyList_GET_SIZE(stack_list);
    if (level>list_length-1)
        level = list_length-1;

    frame_record = PyList_GetItem(stack_list, level);
    if (!frame_record)
        goto done;

    frame = PyTuple_GetItem(frame_record, 0);
    if (!frame)
        goto done;

    f_code = PyObject_GetAttrString(frame, "f_code");
    if (!f_code)
        goto done;

    co_name = PyObject_GetAttrString( f_code, "co_name");
    if (!co_name)
        goto done;

    out = PyString_AsString(co_name);

done:
    if ( PyErr_Occurred())
    {
        print_python_error();
        PyErr_Print();
    }
    Py_XDECREF(co_name);
    Py_XDECREF(f_code);
    Py_XDECREF(stack_list);
    Py_XDECREF(stack_func);
    Py_XDECREF(inspect);
    return out;
}


/* Obtains a string from a Python traceback.
   This is the exact same string as "traceback.print_exc" would return.

   Pass in a Python traceback object (probably obtained from PyErr_Fetch())
   Result is a string which must be free'd using PyMem_Free()
*/
#define TRACEBACK_FETCH_ERROR(what) {errMsg = what; goto done;}

char *PyTraceback_AsString(PyObject *exc_tb)
{
    const char *errMsg = NULL; /* a static that hold a local error message */
    char *result = NULL; /* a valid, allocated result. */
    PyObject *modStringIO = NULL;
    PyObject *modTB = NULL;
    PyObject *obFuncStringIO = NULL;
    PyObject *obStringIO = NULL;
    PyObject *obFuncTB = NULL;
    PyObject *argsTB = NULL;
    PyObject *obResult = NULL;

    /* Import the modules we need - cStringIO and traceback */
    modStringIO = PyImport_ImportModule("cStringIO");
    if (modStringIO==NULL)
        TRACEBACK_FETCH_ERROR("cant import cStringIO\n");

    modTB = PyImport_ImportModule("traceback");
    if (modTB==NULL)
        TRACEBACK_FETCH_ERROR("cant import traceback\n");
    /* Construct a cStringIO object */
    obFuncStringIO = PyObject_GetAttrString(modStringIO, "StringIO");
    if (obFuncStringIO==NULL)
        TRACEBACK_FETCH_ERROR("cant find cStringIO.StringIO\n");
    obStringIO = PyObject_CallObject(obFuncStringIO, NULL);
    if (obStringIO==NULL)
        TRACEBACK_FETCH_ERROR("cStringIO.StringIO() failed\n");
    /* Get the traceback.print_exception function, and call it. */
    obFuncTB = PyObject_GetAttrString(modTB, "print_tb");
    if (obFuncTB==NULL)
        TRACEBACK_FETCH_ERROR("cant find traceback.print_tb\n");

    argsTB = Py_BuildValue("OOO",
            exc_tb  ? exc_tb  : Py_None,
            Py_None,
            obStringIO);
    if (argsTB==NULL)
        TRACEBACK_FETCH_ERROR("cant make print_tb arguments\n");

    obResult = PyObject_CallObject(obFuncTB, argsTB);
    if (obResult==NULL)
        TRACEBACK_FETCH_ERROR("traceback.print_tb() failed\n");
    /* Now call the getvalue() method in the StringIO instance */
    Py_DECREF(obFuncStringIO);
    obFuncStringIO = PyObject_GetAttrString(obStringIO, "getvalue");
    if (obFuncStringIO==NULL)
        TRACEBACK_FETCH_ERROR("cant find getvalue function\n");
    Py_DECREF(obResult);
    obResult = PyObject_CallObject(obFuncStringIO, NULL);
    if (obResult==NULL)
        TRACEBACK_FETCH_ERROR("getvalue() failed.\n");

    /* And it should be a string all ready to go - duplicate it. */
    if (!PyString_Check(obResult))
            TRACEBACK_FETCH_ERROR("getvalue() did not return a string\n");

    { // a temp scope so I can use temp locals.
    char *tempResult = PyString_AsString(obResult);
    result = (char *)PyMem_Malloc(strlen(tempResult)+1);
    if (result==NULL)
        TRACEBACK_FETCH_ERROR("memory error duplicating the traceback string");

    strcpy(result, tempResult);
    } // end of temp scope.
done:
    /* All finished - first see if we encountered an error */
    if (result==NULL && errMsg != NULL) {
        result = (char *)PyMem_Malloc(strlen(errMsg)+1);
        if (result != NULL)
            /* if it does, not much we can do! */
            strcpy(result, errMsg);
    }
    Py_XDECREF(modStringIO);
    Py_XDECREF(modTB);
    Py_XDECREF(obFuncStringIO);
    Py_XDECREF(obStringIO);
    Py_XDECREF(obFuncTB);
    Py_XDECREF(argsTB);
    Py_XDECREF(obResult);
    return result;
}


static void format_exception(std::string &msg,
                             PyObject *exc_typ,
                             PyObject *exc_val,
                             PyObject *exc_tb)
{
    std::ostringstream ostr;
    msg.clear();
    if (!exc_typ)
        return;

    if (exc_tb)
    {
        char *traceback = PyTraceback_AsString(exc_tb);
        if (traceback == NULL)
            ostr << "Can't get the traceback info!";
        else
        {
            ostr << "Traceback (most recent call last):\n" << traceback;
            PyMem_Free(traceback);
        }
    }
    PyObject *temp = PyObject_Str(exc_typ);
   if (temp)
   {
       ostr << PyString_AsString(temp);
       Py_DECREF(temp);
   }
   else
   {
       ostr << "Can't convert exception to a string!";
   }
   ostr << ": ";
   if (exc_val != NULL)
   {
       temp = PyObject_Str(exc_val);
       if (temp)
       {
           ostr << PyString_AsString(temp);
           Py_DECREF(temp);
       }
       else
       {
           ostr << "Can't convert exception value to a string!";
       }
   }
   msg = ostr.str();
}
void print_python_error()
{
    std::string msg;
    PyObject *exc_typ = NULL, *exc_val = NULL, *exc_tb = NULL;
    PyErr_Fetch(&exc_typ, &exc_val, &exc_tb);
    PyErr_NormalizeException(&exc_typ, &exc_val, &exc_tb);
    if (exc_typ)
    {
        format_exception(msg, exc_typ, exc_val, exc_tb);
        if (!msg.empty())
            std::cout << msg << std::endl;
    }
    PyErr_Restore(exc_typ, exc_val, exc_tb);
}

int find_functions_in_module(const char *module_name, std::vector<std::string> &vout)
{
    Py_ssize_t item, length;
    PyObject *inspect=0, *module=0, *dict=0, *keys, *key, *obj, *rc, *isfunc;

    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();

    vout.clear();
    inspect = PyImport_ImportModule("inspect");
    if (!inspect)
    {
        if (PyErr_Occurred())
        {
            print_python_error();
            PyErr_Print();
        }
        PyGILState_Release(gilstate);
        return 0;
    }
    isfunc = PyObject_GetAttrString(inspect, "isfunction");

    gchar *fnam = g_strdup(module_name);
    gchar *p = g_strrstr(fnam, ".py");
    if (p)
        *p = 0;
    module = PyImport_ImportModule(fnam);
    g_free(fnam);
    if (!module)
    {
        if (PyErr_Occurred())
        {
            print_python_error();
            PyErr_Print();
        }
        Py_DECREF(isfunc);
        Py_DECREF(inspect);
        PyGILState_Release(gilstate);
        return 0;
    }

    dict = PyModule_GetDict(module);
    keys = PyDict_Keys(dict);
    length = PyMapping_Length(dict);
    for (item=0;item<length;item++)
    {
        key = PyList_GetItem(keys, item);
        obj = PyDict_GetItem(dict, key);
        rc = PyObject_CallFunctionObjArgs(isfunc, obj,NULL);
        if ( PyInt_AsLong(rc) )
        {
            vout.push_back( PyString_AsString(key));
        }
        Py_DECREF(rc);
    }
    Py_DECREF(keys);
    Py_DECREF(dict);
    Py_DECREF(module);
    Py_DECREF(isfunc);
    Py_DECREF(inspect);
    PyGILState_Release(gilstate);

    return vout.size();
}


static PyObject *py_err_msg(PyObject *self, PyObject *args, PyObject *kwargs)
{
    // TODO: find a way to figure out the name of the calling function
    // and use err_msg_ex
    PyObject *pymsg;
    std::string func_name;
    const char *txt;
    if (!PyArg_ParseTuple(args,"S", &pymsg))
        return 0;

    txt = PyString_AsString(pymsg);
    func_name = get_function_name();
    err_msg_ex(func_name.c_str(), 0, "%s\n", txt);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *py_write_msg(PyObject *self, PyObject *args, PyObject *kwargs)
{
    PyObject *pymsg;
    const char *txt;
    if (!PyArg_ParseTuple(args,"S", &pymsg))
        return 0;

    txt = PyString_AsString(pymsg);
    write_msg("%s\n", txt);

    Py_INCREF(Py_None);
    return Py_None;
}


static PyObject *py_info_msg(PyObject *self, PyObject *args, PyObject *kwargs)
{
    // TODO: find a way to figure out the name of the calling function
    // and use info_msg_ex
    PyObject *pymsg;
    std::string func_name;
    const char *txt;
    if (!PyArg_ParseTuple(args,"S", &pymsg))
        return 0;

    func_name = get_function_name(1);
    txt = PyString_AsString(pymsg);
    info_msg_ex(func_name.c_str(),0, "%s\n", txt);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyMethodDef alibava_functions[] =
{
    {"error_msg",(PyCFunction)py_err_msg, METH_VARARGS,"Prints an error message"},
    {"info_msg",(PyCFunction)py_info_msg, METH_VARARGS,"Prints an information message"},
    {"write_msg",(PyCFunction)py_write_msg, METH_VARARGS,"Prints a message"},
    { NULL, NULL, 0 }
};

int export_c_functions(PyObject *dict, PyMethodDef *methods)
{
    PyMethodDef *ml;


    if (methods==0)
        return 0;

    if (dict==0 || !PyDict_Check(dict))
        return -1;

    for (ml = methods; ml->ml_name != NULL; ml++)
    {
        PyObject *func = PyCFunction_NewEx(ml,0,0);
        if (!func)
        {
            if ( PyErr_Occurred())
            {
                print_python_error();
                PyErr_Print();
            }
            return -2;
        }
        if (PyDict_SetItemString(dict, ml->ml_name, func) != 0)
        {
            Py_DECREF(func);
            return -3;
        }
        Py_DECREF(func);
    }
    return 0;
}

int PythonLoader::export_functions(PyObject *module)
{
    int rc;
    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();
    PyObject *dict = PyModule_GetDict(module);
    rc = export_c_functions(dict, alibava_functions);
    PyGILState_Release(gilstate);
    return rc;
}

PythonLoader::PythonLoader() :
    _initialized(false), _save(0)
{
    std::cout << "Initializing Python" << std::endl;
    Py_Initialize();
    init_python_objects();
    // Get prepared for the threading...
    if (! PyEval_ThreadsInitialized() )
    {
        PyThreadState *py_tstate = NULL;
        PyEval_InitThreads();
        py_tstate = PyGILState_GetThisThreadState();
        PyEval_ReleaseThread(py_tstate);
    }
}

void PythonLoader::initialize(int argc, char **argv)
{
    if (_initialized)
        return;
    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();

    _initialized = true;
    PySys_SetArgv(argc, argv);
    PyRun_SimpleString("import sys\n");
    PyGILState_Release(gilstate);
}

void PythonLoader::show_path()
{
    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();

    std::cout << "System search path:\n" << std::flush;
    PyObject *path = PySys_GetObject("path");
    if ( path )
    {
        for (Py_ssize_t i = 0; i< PyList_Size(path); ++i) {
            PyObject* obj = PyList_GetItem(path, i);
            if (obj != NULL) {
                std::cout << '\t' << PyString_AsString(obj) << std::endl;
            }
        }
    }
    std::cout << std::endl;
    PyGILState_Release(gilstate);
}

void add_python_path(const char *p)
{
    std::ostringstream ostr;
    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();

    PyObject *path = PySys_GetObject("path");
    bool do_add = true;
    if ( path )
    {
        if ( __search_path__.empty() )
        {
            for (Py_ssize_t i = 0; i< PyList_Size(path); ++i)
            {
                PyObject* obj = PyList_GetItem(path, i);
                __search_path__[ PyString_AsString(obj) ] = 1;
            }
        }
        else
        {
            std::map< std::string, int>::iterator ipath = __search_path__.find( p );
            if (ipath==__search_path__.end() )
                __search_path__[ p ] = 1;
            else
                do_add = false;
        }
        if ( do_add )
            PyList_Append(path, PyString_FromString(p));
    }
    if ( PyErr_Occurred() )
    {
        print_python_error();
        PyErr_Print();
    }
    PyGILState_Release(gilstate);
}

void PythonLoader::add_path(const char *p)
{
    add_python_path(p);
}

PythonLoader::~PythonLoader()
{
    std::cout << "Finalizing Python" << std::endl;

    ObjectList::iterator ip;
    for (ip=MethodList.begin();ip!=MethodList.end();ip++)
        Py_DECREF(ip->second);

    for (ip=ModuleList.begin();ip!=ModuleList.end();ip++)
        Py_DECREF(ip->second);

    PyGILState_Ensure();
    Py_Finalize();
}

PythonLoader &PythonLoader::getInstance()
{
    return __the_loader;
}

PyObject *PythonLoader::load_module(const std::string &module, const std::string &method)
{
    PyObject *pymodule, *pymethod;

    // Load the module
    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();

    ObjectList::iterator ip = ModuleList.find(module);
    if (ip==ModuleList.end())
    {
        char *mod_nam = const_cast<char *>(module.c_str());
        pymodule = PyImport_ImportModule(mod_nam);
        if (pymodule==0)
        {
            std::cout << "Could not load " << module << std::endl;
            if ( PyErr_Occurred())
            {
                print_python_error();
                PyErr_Print();
            }
            return 0;
        }
        ModuleList[module] = pymodule;
        Py_INCREF(pymodule);
    }
    else
    {
        pymodule = ip->second;
        Py_INCREF(pymodule);
    }
    PyGILState_Release(gilstate);
    /*
     * Now look for the hook
     */
    std::string name = method + '@' + module;
    ip = MethodList.find(name);
    if (ip==MethodList.end())
    {
        char *method_name = const_cast<char *>(method.c_str());
        gilstate = PyGILState_Ensure();
        pymethod = PyObject_GetAttrString(pymodule, method_name);
        if (!pymethod)
        {
            std::cout << "Could not find " << method << " in " << module << std::endl;
            PyGILState_Release(gilstate);
            return 0;
        }

        if ( !PyCallable_Check(pymethod))
        {
            std::cout << method << " in " << module << " is not a callable function" << std::endl;
            PyGILState_Release(gilstate);
            return 0;
        }
        MethodList[name] = pymethod;
        Py_INCREF(pymethod);
        PyGILState_Release(gilstate);
    }
    else
        pymethod = ip->second;

    return pymethod;
}


PyObject *PythonLoader::execute_function_from_module(const char *module_name,
                                                     const char *function_name,
                                                     bool export_funcs,
                                                     PyObject **module_ptr,
                                                     PyObject **method_ptr)
{
    if (module_ptr)
        *module_ptr = 0;

    if (method_ptr)
        *method_ptr = 0;


    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();

    gchar *fnam = g_strdup(module_name);
    gchar *p = g_strrstr(fnam, ".py");
    if (p)
        *p = 0;

    // load the python module
    PyObject *func = 0;
    PyObject *ret=0;
    PyObject *mName = PyString_FromString(fnam);
    PyObject *module = PyImport_Import( mName );
    g_free(fnam);
    if (!module)
    {
        if (PyErr_Occurred())
        {
            print_python_error();
            PyErr_Print();
        }
        std::cout << "Module " << module_name << " not found." << std::endl;
        show_path();
    }
    else
    {
        PyObject *dict = PyModule_GetDict(module);
        func = PyDict_GetItemString(dict, function_name);
        if (!func)
        {
            std::cout << "Function " << function_name << " not found in " << module_name << std::endl;
            if (PyErr_Occurred())
            {
                print_python_error();
                PyErr_Print();
            }

            Py_DECREF(mName);
            PyGILState_Release(gilstate);
            return 0;
        }
        if (export_funcs)
            export_functions(module);

        ret = PyObject_CallObject(func, 0);
        if (!ret)
        {
            if (PyErr_Occurred())
            {
                print_python_error();
                PyErr_Print();
            }
        }
        if (module_ptr)
            *module_ptr = module;

        if (method_ptr)
            *method_ptr = func;
        Py_DECREF(mName);
    }

    PyGILState_Release(gilstate);
    return ret;
}


void PythonLoader::allow_threads()
{
    _save = PyEval_SaveThread();
}

void PythonLoader::disable_threads()
{
    if (_save)
    {
        PyEval_RestoreThread(_save);
        _save = 0;
    }
}
void PythonLoader::wait(double secs)
{
    struct timeval t;
    double frac;
    frac = fmod(secs, 1.0);
    secs = floor(secs);
    t.tv_sec = (long)secs;
    t.tv_usec = (long)(frac*1000000.0);
    Py_BEGIN_ALLOW_THREADS
    if (select(0, NULL, NULL, NULL, &t) != 0) {
        if (errno!=EINTR)
            std::cout << "Error in python_wait: " << strerror(errno) << std::endl;
    }
    Py_END_ALLOW_THREADS
}

#endif
