/* -*- mode: c++ -*- */
#ifndef USBPORT_H_
#define USBPORT_H_

#include <string>

class Fifo;

// Abstract class defining operations on the USB port
class USBport
{
    protected:
        std::string devname; // name of the device

    public:
        // Creates and opend the port controller
        USBport();
        USBport(const char *fnam);

        virtual ~USBport();

        const std::string &device() const
        {
            return devname;
        }

        /**
         * Returns the error status.
         * 0 means no errors
         * */
        virtual int error() const
        {
            return 0;
        }
        bool operator!() const
        {
            return error()!=0;
        }

        virtual int open(const std::string &fnam)=0;
        virtual int close()=0;
        virtual int read(void *buf, int size) = 0;
        virtual int write(void *buf, int size) = 0;
        virtual int flush() = 0;
        
        /*
         * Sets the buffer size. 
         * This is only useful when the class implements a buffer to store
         * the data.
         */
        virtual void buffer_size(unsigned int ) {}

        /*
         * Returns the data available for reading
         */
        virtual unsigned int data_available(unsigned int size)
        {
            return size;
        }

        /**
         * Waits tmout seconds for new data. 
         * Returns the ammount of data available 
         */
        virtual int wait_for_data(double tmout=1.e20);

        /**
         * Polls for new data arriving.
         * Timeout: if <0 wait for ever
         *          if >0 wait for timeout milliseconds
         *
         *  Returns: 0  if there is data
         *           >0 if the call gave a timeout
         *           <0 an error
         */
        virtual int poll(int timeout)
        {
            return 0;
        }
        
        virtual int fildes() const { return -1; }

        /*
         * This is used to connect to the Fifo
         */
        virtual int setup_fifo(Fifo *);
};

#endif /*USBPORT_H_*/
