//#define DEBUG 1
#define ASCII 1
#include <glib.h>
#ifdef HAVE_CONFIG_H
#ifdef G_OS_WIN32
#include "config_win.h"
#else
#include <config.h>
#endif
#else
#warning "NO config"
#endif

#ifdef HAVE_GIO

#define ALIBAVA_SERVER
#include <gtk/gtk.h>
#include <glib.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef G_OS_WIN32
#include "mingw.h"
#else
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <sys/un.h>
#endif
#include <cstdlib>
#include <cstdio>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <gio/gio.h>
#include <math.h>
#include <csignal>
#include <getopt.h>
#include "alibava-server.h"

#include <iostream>
#include <map>
#include <string>
#include "string_utils.h"
#include "Machine.h"
#include "Alibava.h"
static void start_subprocess();
static std::map<int, struct client_info *> clients;
static GMainLoop *loop=NULL;

TheMachine machine;
extern int __is_gui_session__;

static bool do_run=true;

static void quit(int sg)
{
    info_msg("Have to quit....\n");
    g_main_loop_quit(loop);
}




static void my_log (const gchar *log_domain,
                    GLogLevelFlags log_level,
                    const gchar *message,
                    gpointer user_data)
{
    if (!strcmp(log_domain,"libglade") && strstr(message,"unknown"))
        return;

    printf("MyLogger: %s\n", message);
    //    (*old_log)(log_domain,log_level,message,user_data);
}
static void setup_logger()
{
    g_log_set_handler ("GLib-GObject",
                       (GLogLevelFlags)(G_LOG_LEVEL_MASK |
                                        G_LOG_FLAG_FATAL |
                                        G_LOG_FLAG_RECURSION),
                       my_log, 0);
    g_log_set_handler ("Gtk",
                       (GLogLevelFlags)(G_LOG_LEVEL_MASK |
                                        G_LOG_FLAG_FATAL |
                                        G_LOG_FLAG_RECURSION),
                       my_log, 0);
    g_log_set_handler ("GLib",
                       (GLogLevelFlags)(G_LOG_LEVEL_MASK |
                                        G_LOG_FLAG_FATAL |
                                        G_LOG_FLAG_RECURSION),
                       my_log, 0);
    g_log_set_handler ("libglade",
                       (GLogLevelFlags)(G_LOG_LEVEL_MASK |
                               G_LOG_FLAG_FATAL |
                               G_LOG_FLAG_RECURSION),
                               my_log, 0);

}

static int init_daemon()
{
#ifdef G_OS_UNIX
    pid_t pid;
    /**
     Become a deamon.
     */
    if ( (pid=fork())<0)
        return -1;
    else if (pid!=0)
        exit(0); // parent exits

    // Child continues
    setsid(); // become session leader

    chdir("/"); // change working directory
    umask(0); // clear our file mode creation mask
#endif
    return 0;
}

int main(int argc, char *argv[] )
{
    bool is_daemon = true;
    int  fw_version = 2;
    gchar *daemon_value= (gchar *)"yes";
    std::string value;

    GError *error = NULL;
    GOptionContext *context;
    static GOptionEntry cmd_entries[] =
    {
     {"daemon",'d', 0, G_OPTION_ARG_STRING, &daemon_value, "tells if the server runs in daemon mode or not", "yes|no" },
     {"firmware-version", 'v', 0, G_OPTION_ARG_INT, &fw_version, "The firmaware version", NULL},
     {NULL}
    };

    __is_gui_session__ = 0;


    // Initialize
    g_thread_init(NULL);
    g_type_init();
    //gtk_init(&argc, &argv);

    setup_logger();

    context = g_option_context_new(" - The alibava emulator server");
    g_option_context_add_main_entries (context, cmd_entries, NULL);
    if ( !g_option_context_parse(context, &argc, &argv, &error))
    {
        g_printerr ("%s: %s\n", argv[0], error->message);
        return 1;
    }

    g_printerr("alibava-server\n");

    value = lower_string(daemon_value);
    if (value != "yes" && value!="no")
    {
        std::cout << "argument of 'daemon' should be yes or no" << std::endl;
        return 1;
    }
    if (value=="no")
        is_daemon = false;

    printf("Firmware version %d\n", fw_version);
	machine.set_fw_version(fw_version);
    if ( is_daemon )
    {
        if (init_daemon())
            return 1;
    }

    

    start_subprocess();

    return 0;
}


static GSocket *create_socket(const char *name)
{
    int opt, sockfd;
    GError *error=0;
    GSocketAddress *src_address;

    /**
     * We ignore the name. We abandon unix sockets so that
     * code can be more easily ported
     */
    PDEBUG( "Creating a new socket" );
    GSocket *sckt = g_socket_new(G_SOCKET_FAMILY_IPV4,
                                   G_SOCKET_TYPE_STREAM,
                                   G_SOCKET_PROTOCOL_DEFAULT,
                                   &error);
    if (sckt == NULL)
    {
        g_printerr ("%s: %s\n", "create_socket: ", error->message);
        g_error_free(error);
        return 0;
    }


    opt = 1;
    sockfd = g_socket_get_fd(sckt);
    setsockopt (sockfd, IPPROTO_TCP,
                    TCP_NODELAY, (char *) &opt, sizeof (opt));
    setsockopt (sockfd, SOL_SOCKET,
                    SO_REUSEADDR, (char *) &opt, sizeof (opt));


    PDEBUG("...Done");
    PDEBUG("Now bind the socket");
    src_address = g_inet_socket_address_new (g_inet_address_new_any (G_SOCKET_FAMILY_IPV4),
                                             SERVER_PORT);
    if (!g_socket_bind (sckt, src_address, TRUE, &error))
    {
        g_printerr ("Can't bind socket: %s\n", error->message);
        g_error_free(error);
        return 0;
    }
    g_object_unref (src_address);
    PDEBUG("...Done");
    PDEBUG("Socket Listens");
    if (!g_socket_listen (sckt, &error))
    {
        g_printerr ("Can't listen on socket: %s\n", error->message);
        g_error_free(error);
        return 0;
    }
    PDEBUG("...Done");

    g_socket_set_blocking(sckt, TRUE);

    return sckt;
}

static GSocket *accept_connection(GSocket *S, int *uidptr, int *type)
{
    static int counter=0;
    GError *error = 0;
    GSocket *new_socket;

    PDEBUG("new connection");

//    PDEBUG("waiting for IN condition");
//    if (!g_socket_condition_wait (S, (GIOCondition)(G_IO_IN | G_IO_OUT | G_IO_PRI), NULL, &error))
//    {
//        g_printerr ("condition wait error for %s: %s\n",
//                    "accept_connection",
//                    error->message);
//        exit (1);
//    }


    PDEBUG("...calling accept");
    new_socket = g_socket_accept (S, NULL, &error);
    if (!new_socket)
    {
        g_printerr ("Error accepting socket: %s\n",
                    error->message);
        g_error_free(error);
        return 0;
    }

    // non blocking: we can also use g_socket_set_timeout
    g_socket_set_blocking (new_socket, TRUE);

    if (uidptr!=NULL)
        *uidptr = counter++;

    PDEBUG("...uidptr: %d", *uidptr);

    // get the tyP/pe
    if (type != NULL)
        *type = *uidptr % 2 ? 0 : 1;

    PDEBUG("...type: %d", *type);

    int opt_val = 1;
    setsockopt( g_socket_get_fd(new_socket),
                IPPROTO_TCP,
                TCP_NODELAY, (char *)&opt_val,
                sizeof(int));

    return new_socket;
}

static gboolean server_client_io(GIOChannel * channel, GIOCondition condition, struct client_info *data)
{

    if( condition != (G_IO_IN) && condition != G_IO_HUP )
        return true;

    PDEBUG("client IO: %d",condition);

    if (condition == G_IO_HUP)
    {
        printf("file HUP\n");
        exit(1);
    }
    else if (condition == G_IO_IN)
    {
        PDEBUG("client wants to write");
        int rc = machine.process_state(data->io_chan);
        if (rc<0)
        {
            PDEBUG("There has been an error in Machine::process_state");
            // There has been an error, give it back
        }
        PDEBUG("Command processed");
    }
    return TRUE;
}

static void remove_client(struct client_info *data)
{
    GError *error = 0;
    gboolean rc;
    PDEBUG("removing client");

    g_mutex_free( data->mux );
    g_socket_close( data->fd, &error );
    if ( error )
    {
        g_printerr ("Cannot close the socket: %s\n",
                            error->message);
        g_error_free(error);
        error = 0;
    }

    g_socket_close( data->cfd, &error );
    if ( error )
    {
        g_printerr ("Cannot close the opmmand socket: %s\n",
                            error->message);
        g_error_free(error);
    }

    do
    {
        rc = g_source_remove_by_user_data(data);
    } while (rc);

    clients.erase(data->uid);
    do_run = false;
}

static gboolean server_client_reads(struct client_info *data)
{
/*
    if (machine.mstate()==TheMachine::ReadingData)
    {
        //        printf("server_client_reads\n");
        if (machine.send_data_block(data->fd, data->size)<0)
        {
            machine.set_to_wait(data->fd);
        }
    }
    */
    return FALSE;
}

static gboolean server_ioctl(GIOChannel * channel, GIOCondition condition, struct client_info *data)
{
    GError *err = 0;
    char type;
    int nr, dir, sz, rc = 0;
    int die = 0;
    gsize nread;
    size_t size;
    GIOStatus status;

    if (condition == G_IO_HUP)
    {
        printf("control HUP\n");
        exit(1);
    }

    // first read the command
    status = g_io_channel_read_chars(channel,(gchar *)&data->command, sizeof(int), &nread, &err);
    if (err)
    {
        g_printerr ("server_ioctl: error getting command %s\n",
                                    err->message);
        g_error_free(err);
        return TRUE;
    }
    PDEBUG("Receiving command: %x rc=%d", data->command, nread);
    type = _IOC_TYPE (data->command);
    nr = _IOC_NR (data->command);
    dir = _IOC_DIR (data->command);
    sz = _IOC_SIZE (data->command);

    switch (nr)
    {
    case RESET_NR:
        PDEBUG ("I have to RESET");
        break;

    case DIE_NR:
        PDEBUG ("I have to quit");
        die = 1;
        break;

    case ERROR_NR:
        rc = machine.error();
        break;

    case DO_READ_NR:
        PDEBUG("DO_READ");
        for (nread=0; nread == 0; )
            status = g_io_channel_read_chars(channel,(gchar *)&size, sz, &nread, &err);

        if ( nread < 0 )
        {
            if ( err ) {
                err_msg(_("Error getting read size: %s\n"), err->message);
                g_error_free(err);
                err = 0;
            }
            else
                err_msg(_("Error getting read size\n"));

        }
        PDEBUG (_("do_read size=%d (%d) rc=%d"), size, sz, nread);
        data->size = size;
        g_idle_add((GSourceFunc) server_client_reads, data);
        break;
    default:
        PDEBUG ("ioctl Unknown command %x", data->command);
        rc = -1;
        break;
    }
    // write back the return code
    PDEBUG ("ioctl sends %d as rc", rc);
    g_socket_send(data->cfd, (gchar *)&rc, sizeof(int), 0, &err);
    if (err)
    {
        exit(1);
        //err_msg(_("Cannot send back return code: %s\n"), err->message);
        //g_error_free(err);
        //err = 0;
    }
    if (die)
    {
        // we do not wait for a return code when closing
        PDEBUG ("removing clients");
        remove_client(data);
        do_run = false;
    }

    return TRUE;
}

static gpointer server_accepts(gpointer ptr)
{
    int type;
    int uid, duid;
    unsigned long word = 0xbebecafe;
    struct client_info *data;

    GSocket *sckt = (GSocket *)ptr;
    GSocket *sss;
    GIOChannel *chan;
    GError *error = 0;

    while(do_run)
    {
        PDEBUG("new connection\n");

        sss = accept_connection(sckt, &uid, &type);
        if (!do_run)
        {
            g_socket_close(sss, &error);
            PDEBUG("have to quit");
            break;
        }
        PDEBUG("...accepted: %x", sss);
        // Check if we have this client in the list
        duid = 2*(uid/2);

        std::map<int, struct client_info *>::iterator ip = clients.find(duid);
        if (ip==clients.end())
        {
            PDEBUG("...new client");
            data = g_new0(struct client_info,1);
            data->uid = duid;
            data->mux = g_mutex_new();
            clients[duid]=data;
        }
        else
        {
            PDEBUG("...existing client");
            data = ip->second;
        }

        if (type)
        { // this is the ioctl channel
            PDEBUG("...creating control channel");
            data->cfd=sss;
#ifdef G_OS_WIN32
            chan = g_io_channel_win32_new_socket( g_socket_get_fd(sss));
#else
            chan = g_io_channel_unix_new( g_socket_get_fd(sss));
#endif
            data->control = g_io_add_watch(chan, (GIOCondition)(G_IO_IN | G_IO_HUP), (GIOFunc)server_ioctl, data);
            data->control_chan = chan;
        }
        else
        {
            PDEBUG("...creating data channel");
            data->fd = sss;
#ifdef G_OS_WIN32
            chan = g_io_channel_win32_new_socket( g_socket_get_fd(sss));
#else
            chan = g_io_channel_unix_new( g_socket_get_fd(sss));
#endif
            data->io = g_io_add_watch(chan, (GIOCondition)(G_IO_IN | G_IO_HUP), (GIOFunc)server_client_io, data);
            data->io_chan = chan;
        }

        g_io_channel_set_encoding(chan, 0, &error);
        g_io_channel_set_buffered(chan, FALSE);
        g_io_channel_ref(chan);

        error = 0;
        g_socket_send(sss, (gchar *)&word, sizeof(unsigned long), 0, &error);
        if (error)
        {
            err_msg(_("Cannot send handshake word: %s\n"), error->message);
            g_error_free(error);
            error = 0;
        }
    }
    return NULL;
}

/*
 Starts the daemon process. It creates a socket that will communicate
 the user application with the server
 */
static void start_subprocess()
{
    GSocket *sckt;
    GError *error=0;
    
    PDEBUG("Server starting");
    

    PDEBUG("Creating socket");
    sckt = create_socket("/tmp/alibava.socket");
    if (!sckt)
        exit(1);
    PDEBUG("...socket fildes: %x", sckt);


    // init the random number generator
    srand(time(0));


    /** TODO: Make things easy (for windows), since we will not have
     *  multiple users... 
     * Just create an extra thread which calls accept (we limit listen 
     * to 2 connections) andd the watchers to the created streams.
     *  
     */
    // init the list of clients
    clients.clear();
    loop = g_main_loop_new(NULL, FALSE);
    g_thread_create(server_accepts,(gpointer)sckt, FALSE,NULL);

    signal(SIGINT, quit);
    signal(SIGTERM, quit);

    PDEBUG("Start loop");
    g_main_loop_run(loop);
    g_socket_close(sckt, &error);

    return;
}
#endif
